from __future__ import unicode_literals

import cgi
import codecs
import logging
import sys
from io import BytesIO
from threading import Lock

from django import http
from django.core import signals
from django.core.handlers import base
from django.core.urlresolvers import set_script_prefix
from django.utils import six
from django.utils.functional import cached_property

from .exceptions import ResponseLater as ResponseLaterOuter

logger = logging.getLogger('django.request')


class AsgiRequest(http.HttpRequest):
    """
    Custom request subclass that decodes from an ASGI-standard request
    dict, and wraps request body handling.
    """

    ResponseLater = ResponseLaterOuter

    def __init__(self, message):
        self.message = message
        self.reply_channel = self.message.reply_channel
        self._content_length = 0
        self._post_parse_error = False
        self.resolver_match = None
        # Path info
        self.path = self.message['path'].decode("ascii")
        self.script_name = self.message.get('root_path', b'')
        if self.script_name:
            # TODO: Better is-prefix checking, slash handling?
            self.path_info = self.path[len(self.script_name):]
        else:
            self.path_info = self.path
        # HTTP basics
        self.method = self.message['method'].upper()
        self.META = {
            "REQUEST_METHOD": self.method,
            "QUERY_STRING": self.message.get('query_string', b'').decode("ascii"),
            "SCRIPT_NAME": self.script_name.decode("ascii"),
            # Old code will need these for a while
            "wsgi.multithread": True,
            "wsgi.multiprocess": True,
        }
        if self.message.get('client', None):
            self.META['REMOTE_ADDR'] = self.message['client'][0]
            self.META['REMOTE_HOST'] = self.META['REMOTE_ADDR']
            self.META['REMOTE_PORT'] = self.message['client'][1]
        if self.message.get('server', None):
            self.META['SERVER_NAME'] = self.message['server'][0]
            self.META['SERVER_PORT'] = self.message['server'][1]
        # Headers go into META
        for name, value in self.message.get('headers', {}).items():
            if name == "content-length":
                corrected_name = "CONTENT_LENGTH"
            elif name == "content-type":
                corrected_name = "CONTENT_TYPE"
            else:
                corrected_name = 'HTTP_%s' % name.upper().replace("-", "_")
            # HTTPbis say only ASCII chars are allowed in headers
            self.META[corrected_name] = value.decode("ascii")
        # Pull out request encoding if we find it
        if "CONTENT_TYPE" in self.META:
            _, content_params = cgi.parse_header(self.META["CONTENT_TYPE"])
            if 'charset' in content_params:
                try:
                    codecs.lookup(content_params['charset'])
                except LookupError:
                    pass
                else:
                    self.encoding = content_params['charset']
        # Pull out content length info
        if self.META.get('CONTENT_LENGTH', None):
            try:
                self._content_length = int(self.META['CONTENT_LENGTH'])
            except (ValueError, TypeError):
                pass
        # Body handling
        self._body = message.get("body", b"")
        if message.get("body_channel", None):
            while True:
                # Get the next chunk from the request body channel
                chunk = None
                while chunk is None:
                    _, chunk = message.channel_layer.receive_many(
                        [message['body_channel']],
                        block=True,
                    )
                # Add content to body
                self._body += chunk.get("content", "")
                # Exit loop if this was the last
                if not chunk.get("more_content", False):
                    break
        assert isinstance(self._body, six.binary_type), "Body is not bytes"
        # Add a stream-a-like for the body
        self._stream = BytesIO(self._body)
        # Other bits
        self.resolver_match = None

    @cached_property
    def GET(self):
        return http.QueryDict(
            self.message.get('query_string', ''),
            encoding=self._encoding,
        )

    def _get_post(self):
        if not hasattr(self, '_post'):
            self._read_started = False
            self._load_post_and_files()
        return self._post

    def _set_post(self, post):
        self._post = post

    def _get_files(self):
        if not hasattr(self, '_files'):
            self._load_post_and_files()
        return self._files

    POST = property(_get_post, _set_post)
    FILES = property(_get_files)

    @cached_property
    def COOKIES(self):
        return http.parse_cookie(self.META.get('HTTP_COOKIE', ''))


class AsgiHandler(base.BaseHandler):
    """
    Handler for ASGI requests for the view system only (it will have got here
    after traversing the dispatch-by-channel-name system, which decides it's
    a HTTP request)
    """

    initLock = Lock()
    request_class = AsgiRequest

    # Size to chunk response bodies into for multiple response messages
    chunk_size = 512 * 1024

    def __call__(self, message):
        # Set up middleware if needed. We couldn't do this earlier, because
        # settings weren't available.
        if self._request_middleware is None:
            with self.initLock:
                # Check that middleware is still uninitialized.
                if self._request_middleware is None:
                    self.load_middleware()
        # Set script prefix from message root_path
        set_script_prefix(message.get('root_path', ''))
        signals.request_started.send(sender=self.__class__, message=message)
        # Run request through view system
        try:
            request = self.request_class(message)
        except UnicodeDecodeError:
            logger.warning(
                'Bad Request (UnicodeDecodeError)',
                exc_info=sys.exc_info(),
                extra={
                    'status_code': 400,
                }
            )
            response = http.HttpResponseBadRequest()
        else:
            try:
                response = self.get_response(request)
            except AsgiRequest.ResponseLater:
                # The view has promised something else
                # will send a response at a later time
                return
        # Transform response into messages, which we yield back to caller
        for message in self.encode_response(response):
            # TODO: file_to_stream
            yield message

    def process_exception_by_middleware(self, exception, request):
        """
        Catches ResponseLater and re-raises it, else tries to delegate
        to middleware exception handling.
        """
        if isinstance(exception, AsgiRequest.ResponseLater):
            raise
        else:
            return super(AsgiHandler, self).process_exception_by_middleware(exception, request)

    def handle_uncaught_exception(self, request, resolver, exc_info):
        """
        Propagates ResponseLater up into the higher handler method,
        processes everything else
        """
        if issubclass(exc_info[0], AsgiRequest.ResponseLater):
            raise
        return super(AsgiHandler, self).handle_uncaught_exception(request, resolver, exc_info)

    @classmethod
    def encode_response(cls, response):
        """
        Encodes a Django HTTP response into ASGI http.response message(s).
        """
        # Collect cookies into headers.
        # Note that we have to preserve header case as there are some non-RFC
        # compliant clients that want things like Content-Type correct. Ugh.
        response_headers = []
        for header, value in response.items():
            if isinstance(header, six.binary_type):
                header = header.decode("latin1")
            if isinstance(value, six.text_type):
                value = value.encode("latin1")
            response_headers.append(
                (
                    six.text_type(header),
                    six.binary_type(value),
                )
            )
        for c in response.cookies.values():
            response_headers.append(
                (
                    'Set-Cookie',
                    c.output(header='').encode("ascii"),
                )
            )
        # Make initial response message
        message = {
            "status": response.status_code,
            "status_text": response.reason_phrase.encode("ascii"),
            "headers": response_headers,
        }
        # Streaming responses need to be pinned to their iterator
        if response.streaming:
            for part in response.streaming_content:
                for chunk, more in cls.chunk_bytes(part):
                    message['content'] = chunk
                    # We ignore "more" as there may be more parts; instead,
                    # we use an empty final closing message with False.
                    message['more_content'] = True
                    yield message
                    message = {}
            # Final closing message
            yield {
                "more_content": False,
            }
        # Other responses just need chunking
        else:
            # Yield chunks of response
            for chunk, last in cls.chunk_bytes(response.content):
                message['content'] = chunk
                message['more_content'] = not last
                yield message
                message = {}

    @classmethod
    def chunk_bytes(cls, data):
        """
        Chunks some data into chunks based on the current ASGI channel layer's
        message size and reasonable defaults.

        Yields (chunk, last_chunk) tuples.
        """
        position = 0
        if not data:
            yield data, True
            return
        while position < len(data):
            yield (
                data[position:position + cls.chunk_size],
                (position + cls.chunk_size) >= len(data),
            )
            position += cls.chunk_size


class ViewConsumer(object):
    """
    Dispatches channel HTTP requests into django's URL/View system.
    """

    handler_class = AsgiHandler

    def __init__(self):
        self.handler = self.handler_class()

    def __call__(self, message):
        for reply_message in self.handler(message):
            message.reply_channel.send(reply_message)



from __future__ import unicode_literals
from asgiref.conformance import make_tests
from ..database_layer import DatabaseChannelLayer

channel_layer = DatabaseChannelLayer(expiry=1)
DatabaseLayerTests = make_tests(channel_layer, expiry_delay=2.1)



from __future__ import unicode_literals

import importlib

from django.core.exceptions import ImproperlyConfigured
from django.utils import six

from .handler import ViewConsumer
from .utils import name_that_thing


class ConsumerRegistry(object):
    """
    Manages the available consumers in the project and which channels they
    listen to.

    Generally this is attached to a backend instance as ".registry"
    """

    def __init__(self, routing=None):
        self.consumers = {}
        # Initialise with any routing that was passed in
        if routing:
            # If the routing was a string, import it
            if isinstance(routing, six.string_types):
                module_name, variable_name = routing.rsplit(".", 1)
                try:
                    routing = getattr(importlib.import_module(module_name), variable_name)
                except (ImportError, AttributeError) as e:
                    raise ImproperlyConfigured("Cannot import channel routing %r: %s" % (routing, e))
            # Load consumers into us
            for channel, handler in routing.items():
                self.add_consumer(handler, [channel])

    def add_consumer(self, consumer, channels):
        # Upconvert if you just pass in a string for channels
        if isinstance(channels, six.string_types):
            channels = [channels]
        # Make sure all channels are byte strings
        channels = [
            channel.decode("ascii") if isinstance(channel, six.binary_type) else channel
            for channel in channels
        ]
        # Import any consumer referenced as string
        if isinstance(consumer, six.string_types):
            module_name, variable_name = consumer.rsplit(".", 1)
            try:
                consumer = getattr(importlib.import_module(module_name), variable_name)
            except (ImportError, AttributeError):
                raise ImproperlyConfigured("Cannot import consumer %r" % consumer)
        # Register on each channel, checking it's unique
        for channel in channels:
            if channel in self.consumers:
                raise ValueError("Cannot register consumer %s - channel %r already consumed by %s" % (
                    name_that_thing(consumer),
                    channel,
                    name_that_thing(self.consumers[channel]),
                ))
            self.consumers[channel] = consumer

    def all_channel_names(self):
        return self.consumers.keys()

    def consumer_for_channel(self, channel):
        try:
            return self.consumers[channel]
        except KeyError:
            return None

    def check_default(self, http_consumer=None):
        """
        Checks to see if default handlers need to be registered
        for channels, and adds them if they need to be.
        """
        if not self.consumer_for_channel("http.request"):
            self.add_consumer(http_consumer or ViewConsumer(), ["http.request"])



import base64
import datetime
import json
import random
import string
import time

from django.apps.registry import Apps
from django.db import DEFAULT_DB_ALIAS, connections, models, transaction
from django.db.utils import OperationalError
from django.utils import six
from django.utils.functional import cached_property
from django.utils.timezone import now


class DatabaseChannelLayer(object):
    """
    ORM-backed ASGI channel layer.

    For development use only; it will span multiple processes fine,
    but it's going to be pretty bad at throughput. If you're reading this and
    running it in production, PLEASE STOP.

    Also uses JSON for serialization, as we don't want to make Django depend
    on msgpack for the built-in backend. The JSON format uses \uffff as first
    character to signify a b64 byte string rather than a text string. Ugly, but
    it's not a valid Unicode character, so it should be safe enough.
    """

    def __init__(self, db_alias=DEFAULT_DB_ALIAS, expiry=60):
        self.expiry = expiry
        self.db_alias = db_alias

    # ASGI API

    extensions = ["groups", "flush"]

    def send(self, channel, message):
        # Typecheck
        assert isinstance(message, dict), "message is not a dict"
        assert isinstance(channel, six.text_type), "%s is not unicode" % channel
        # Write message to messages table
        self.channel_model.objects.create(
            channel=channel,
            content=self.serialize(message),
            expiry=now() + datetime.timedelta(seconds=self.expiry)
        )

    def receive_many(self, channels, block=False):
        if not channels:
            return None, None
        assert all(isinstance(channel, six.text_type) for channel in channels)
        # Shuffle channels
        channels = list(channels)
        random.shuffle(channels)
        # Clean out expired messages
        self._clean_expired()
        # Get a message from one of our channels
        while True:
            try:
                with transaction.atomic():
                    message = self.channel_model.objects.filter(channel__in=channels).order_by("id").first()
                    if message:
                        self.channel_model.objects.filter(pk=message.pk).delete()
                        return message.channel, self.deserialize(message.content)
            except OperationalError:
                # The database is probably trying to prevent a deadlock
                time.sleep(0.1)
                continue
            if block:
                time.sleep(1)
            else:
                return None, None

    def new_channel(self, pattern):
        assert isinstance(pattern, six.text_type)
        # Keep making channel names till one isn't present.
        while True:
            random_string = "".join(random.choice(string.ascii_letters) for i in range(8))
            new_name = pattern.replace("?", random_string)
            if not self.channel_model.objects.filter(channel=new_name).exists():
                return new_name

    # ASGI Group extension

    def group_add(self, group, channel):
        """
        Adds the channel to the named group for at least 'expiry'
        seconds (expiry defaults to message expiry if not provided).
        """
        self.group_model.objects.update_or_create(
            group=group,
            channel=channel,
        )

    def group_discard(self, group, channel):
        """
        Removes the channel from the named group if it is in the group;
        does nothing otherwise (does not error)
        """
        self.group_model.objects.filter(group=group, channel=channel).delete()

    def send_group(self, group, message):
        """
        Sends a message to the entire group.
        """
        self._clean_expired()
        for channel in self.group_model.objects.filter(group=group).values_list("channel", flat=True):
            self.send(channel, message)

    # ASGI Flush extension

    def flush(self):
        self.channel_model.objects.all().delete()
        self.group_model.objects.all().delete()

    # Serialization

    def serialize(self, message):
        return AsgiJsonEncoder().encode(message)

    def deserialize(self, message):
        return AsgiJsonDecoder().decode(message)

    # Database state mgmt

    @property
    def connection(self):
        """
        Returns the correct connection for the current thread.
        """
        return connections[self.db_alias]

    @cached_property
    def channel_model(self):
        """
        Initialises a new model to store messages; not done as part of a
        models.py as we don't want to make it for most installs.
        """
        # Make the model class
        class Message(models.Model):
            # We assume an autoincrementing PK for message order
            channel = models.CharField(max_length=200, db_index=True)
            content = models.TextField()
            expiry = models.DateTimeField(db_index=True)

            class Meta:
                apps = Apps()
                app_label = "channels"
                db_table = "django_channels"
        # Ensure its table exists
        if Message._meta.db_table not in self.connection.introspection.table_names(self.connection.cursor()):
            with self.connection.schema_editor() as editor:
                editor.create_model(Message)
        return Message

    @cached_property
    def group_model(self):
        """
        Initialises a new model to store groups; not done as part of a
        models.py as we don't want to make it for most installs.
        """
        # Make the model class
        class Group(models.Model):
            group = models.CharField(max_length=200)
            channel = models.CharField(max_length=200)

            class Meta:
                apps = Apps()
                app_label = "channels"
                db_table = "django_channel_groups"
                unique_together = [["group", "channel"]]
        # Ensure its table exists
        if Group._meta.db_table not in self.connection.introspection.table_names(self.connection.cursor()):
            with self.connection.schema_editor() as editor:
                editor.create_model(Group)
        return Group

    def _clean_expired(self):
        """
        Cleans out expired groups and messages.
        """
        # Include a 1-second grace period for clock sync drift
        target = now() - datetime.timedelta(seconds=1)
        # First, go through old messages and pick out channels that got expired
        old_messages = self.channel_model.objects.filter(expiry__lt=target)
        channels_to_ungroup = old_messages.values_list("channel", flat=True).distinct()
        old_messages.delete()
        # Now, remove channel membership from channels that expired
        self.group_model.objects.filter(channel__in=channels_to_ungroup).delete()

    def __str__(self):
        return "%s(alias=%s)" % (self.__class__.__name__, self.connection.alias)


class AsgiJsonEncoder(json.JSONEncoder):
    """
    Special encoder that transforms bytestrings into unicode strings
    prefixed with u+ffff
    """

    def transform(self, o):
        if isinstance(o, (list, tuple)):
            return [self.transform(x) for x in o]
        elif isinstance(o, dict):
            return {
                self.transform(k): self.transform(v)
                for k, v in o.items()
            }
        elif isinstance(o, six.binary_type):
            return u"\uffff" + base64.b64encode(o).decode("ascii")
        else:
            return o

    def encode(self, o):
        return super(AsgiJsonEncoder, self).encode(self.transform(o))


class AsgiJsonDecoder(json.JSONDecoder):
    """
    Special encoder that transforms bytestrings into unicode strings
    prefixed with u+ffff
    """

    def transform(self, o):
        if isinstance(o, (list, tuple)):
            return [self.transform(x) for x in o]
        elif isinstance(o, dict):
            return {
                self.transform(k): self.transform(v)
                for k, v in o.items()
            }
        elif isinstance(o, six.text_type) and o and o[0] == u"\uffff":
            return base64.b64decode(o[1:].encode("ascii"))
        else:
            return o

    def decode(self, o):
        return self.transform(super(AsgiJsonDecoder, self).decode(o))



from __future__ import unicode_literals

import logging
import time

from .exceptions import ConsumeLater
from .message import Message
from .utils import name_that_thing

logger = logging.getLogger('django.channels')


class Worker(object):
    """
    A "worker" process that continually looks for available messages to run
    and runs their consumers.
    """

    def __init__(self, channel_layer, callback=None):
        self.channel_layer = channel_layer
        self.callback = callback

    def run(self):
        """
        Tries to continually dispatch messages to consumers.
        """
        channels = self.channel_layer.registry.all_channel_names()
        while True:
            channel, content = self.channel_layer.receive_many(channels, block=True)
            # If no message, stall a little to avoid busy-looping then continue
            if channel is None:
                time.sleep(0.01)
                continue
            # Create message wrapper
            logger.debug("Worker got message on %s: repl %s", channel, content.get("reply_channel", "none"))
            message = Message(
                content=content,
                channel_name=channel,
                channel_layer=self.channel_layer,
            )
            # Handle the message
            consumer = self.channel_layer.registry.consumer_for_channel(channel)
            if self.callback:
                self.callback(channel, message)
            try:
                consumer(message)
            except ConsumeLater:
                self.channel_layer.send(channel, content)
            except:
                logger.exception("Error processing message with consumer %s:", name_that_thing(consumer))
